; Tape <=> serial interface for the Casio FX-702P programmable calculator,
; also supports printing through an optional parallel port.
;
; Xtal frequency 9.8304MHz
;
; Serial port communication protocol: 9600bps, no parity, Xon/Xoff flow control
;
; Data format:
; The 12-bit data word is split in two 6-bit chunks, then added an offset 0x30
; to make them printable characters. The least significant character is sent
; first.
; Characters outside the 0x30-0x6F range are allowed between a pair of valid
; characters. They can be used to synchronize the data stream (to mark the
; least significant character of the pair).

.include "2313def.inc"

; port B
.equ	DATA4		=7	;HD43110 Data
.equ	PRN_ENABLE	=6	;high level enables the printer support,
				;bit number chosen intentionally
.equ	TAPE_ENABLE	=5	;high level enables the tape support
.equ	DATA1		=4
.equ	OP		=3
.equ	CE1		=2
.equ	CE2		=1
.equ	KEY		=0

; port D
.equ	LED		=6
.equ	SDA		=5
.equ	SCL		=4
.equ	SP		=3	;HD43110 Clock, INT1
.equ	CE3		=2	;HD43110 Chip Enable, INT0
.equ	TXD		=1
.equ	RXD		=0

; Optional parallel printer port using I2C bus expanders PCF8574, data
; port on address $42, control and status port on address $40.
; parallel port control and status bits assignments:
.equ	PAR_ACK		=0	;status
.equ	PAR_BUSY	=1	;status
.equ	PAR_PE		=2	;status
.equ	PAR_SLCT	=3	;status
.equ	PAR_SLCTIN	=4	;control
.equ	PAR_INIT	=5	;control
.equ	PAR_ERROR	=6	;status
.equ	PAR_STROBE	=7	;control


; R0 is reserved for the LPM instruction,
; also occasionally as a temporary variable

.def	PrnBusy		=r4	;copy of the printer Busy status signal
				;on bit DATA4, all remaining bits cleared
.def	PrnData		=r5	;data printed
.def	PrnControl	=r6	;printer control data, determines the
				;printer mode
.def	PrnCtrlCopy	=r7	;copy of the PrnControl variable

.def	Const0x80	=r8	;constant, fixed value 0x80

.def	FlowContrChar	=r9	;sent or received flow control character
.def	TapeCounter	=r10	;counter of saved data words,
				;counter of loaded characters

; 16 data bits transferred to the printer (data or control word)
.def	PrnWordLow	=r11
.def	PrnWordHigh	=r12

; 16 data bits transferred between the calculator and the tape interface
.def	DataWordLow	=r13
.def	DataWordHigh	=r14

; variables used exclusively in the IRQ1 interrupt handle routine
.def	IntSregSave	=r15	;stores SREG
.def	IntTemp		=r16	;temporary variable 

.def	Flags		=r17	;tape interface state
; meaning of individual bits:
.equ	TAPE_BUSY	=7	;SAVE mode: bit set after receiving data
				;from the calculator, cleared after sending
				;them through the serial port
				;LOAD mode: bit cleared when data was
				;retrieved from the buffer,
				;set after sending to the calculator,
				;bit numbers of TAPE_BUSY and DATA4
				;intentionally matched
.equ	FORCE_BUSY	=6	;set when Timer1 counting, forces the tape
				;interface to respond "not ready",
				;bit number selected intentionally
.equ	TAPE_LOAD	=5	;set when LOAD, clr when SAVE
.equ	TAPE_DATA_MODE	=4	;set when data transfer
.equ	TAPE_STATUS_MODE =3	;set when status read
.equ	TAPE_ACTIVE	=2	;set when tape communication
.equ	TAPE_START	=1	;set when begin of loading/saving
.equ	CMD_RECEIVED	=0	;set when tape or printer command received

.def	WrPortB		=r18	;data sent to port B
.def	CmdCode		=r19	;command received from the calculator
.def	Counter		=r20	;local general purpose counter
.def	Temp1		=r21	;local temporary variable
.def	Skipped		=r22	;counter of skipped data words after the
				;"open for read" command
.def	DataByte	=r23	;data transferred through the I2C bus,
				;data sent to the calculator's bus

.def	PrnFlags	=r24
; meaning of individual bits:
.equ	PRN_DATA_MODE	=7	;set when data/control transfer
.equ	PRN_CTRL_MODE	=6	;set when transferred data is a control byte
.equ	PRN_STATUS_MODE	=5	;set when status read
.equ	PRN_PRES_MODE	=4	;set when presence signal read
.equ	PRN_DATA_RDY	=3	;set when PrnData contains valid byte
.equ	PRN_CTRL_RDY	=2	;set when PrnControl contains valid byte
.equ	PRN_BUF_FULL	=1	;set after filling up the PrnBuffer, cleared
				;after printing all data out of the PrnBuffer

; X register is the write pointer to the data transfer buffer
; Y register is the read pointer to the data transfer buffer
; Z register is the printer LIFO pointer

.equ	TapeBuffer	=0x60	;circular buffer in the RAM
.equ	TAPEBUFSIZE	=0x50	;size of the buffer

.equ	PrnBuffer	=0xB0	;LIFO buffer used for printing, grows upwards
.equ	PRNBUFSIZE	=0x10	;size of the buffer

; stack area: $C0 to $DF

; The flow control character codes
.equ	XON		=0x11
.equ	XOFF		=0x13


.org	0x0000

		rjmp	start		;Reset Handle
		rjmp	ext_int0	;IRQ0 Handle
		rjmp	ext_int1	;IRQ1 Handle
		reti			;Timer1 capture Handle
		reti			;Timer1 compare Handle
		reti			;Timer1 overflow Handle
		reti			;Timer0 overflow Handle
		reti			;UART RX Complete Handle
		reti			;UDR Empty Handle
		reti			;UART TX Complete Handle
		reti			;Analog Comparator Handle

; Casio to ASCII character conversion tables, codes that don't have their
; ASCII equivalents are marked with ~, values above 0x7F are assigned to
; codes printed as a pair of characters.
; WARNING: the IAR assembler doesn't seem to like certain characters in a
; string. The semicolon must be coded as 0x3B, otherwise it will be treated
; as a comment.
fx702toascii:
.db		":~?!~<>~()AR~~~ "
.db		"~?~", 0x27, 0x22, "#$", 0x3B, ":,>", 0x80, '=', 0x82, '<', 0x84
.db		"+-*/^v!", 0x22, ")@&~(~~~"
.db		"0123456789.", 0x86, "DL-E"
.db		"ABCDEFGHIJKLMNOP"
.db		"QRSTUVWXYZx:+~_~"
.db		"abcdefghijklmnop"
.db		"qrstuvwxyz~~><% "

fx602toascii:
.db		"~~~~~~~~~~~~~~~ "
.db		"abcdefghijklmnop"
.db		"qrstuvwxyz*/><% "
.db		"ABCDEFGHIJKLMNOP"
.db		"QRSTUVWXYZx:+-=~"
.db		":~?!~<>~(),", 0x86, 'E', 0x3B, "#$"
.db		"+-*/^v!", 0x22, ")@&~!?~p"
.db		"0123456789.", 0x22, "'~- "

fx602num:
.db		"0123456789.~~~- "

casio_strings:
.db		">="		;code 0x80
.db		"<="		;code 0x82
.db		"<>"		;code 0x84
.db		"PI"		;code 0x86
.db		0x0D, 0x0A	;code 0x88


; IRQ1 handle routine, used to handle the calculator's bus traffic,
; called at the falling edge of every SP pulse
ext_int1:	sbis	PIND,CE3
		reti
		in	IntSregSave,SREG
		sbis	PINB,OP
		rjmp	bus_cmd

; OP=1 - data transfer
		sbrs	Flags,TAPE_DATA_MODE
		rjmp	tapei_stat
; tape data transfer
		sbrs	Flags,TAPE_LOAD
		rjmp	tapei_save

; tape LOAD mode
		out	PORTB,WrPortB
		sbi	DDRB,DATA4
		lsr	DataWordHigh
		ror	DataWordLow
		ldi	WrPortB,(((1<<DATA1)+(1<<CE1)+(1<<CE2)+(1<<KEY))<<1)
		ror	WrPortB
		rjmp	prni_transfer

; tape SAVE mode
tapei_save:	lsr	DataWordHigh
		ror	DataWordLow
		sbis	PINB,DATA4
		or	DataWordHigh,Const0x80
		rjmp	prni_transfer

; read tape status
tapei_stat:	sbrs	Flags,TAPE_STATUS_MODE
		rjmp	prni_transfer
		mov	IntTemp,Flags
		lsl	IntTemp
		or	IntTemp,Flags		;bit7 = TAPE_BUSY | FORCE_BUSY
		andi	IntTemp,(1<<TAPE_BUSY)	;copy the TAPE_BUSY flag to
						;the IntTemp,DATA4
		ori	IntTemp,(1<<DATA1)+(1<<CE1)+(1<<CE2)+(1<<KEY)
		out	PORTB,IntTemp
		sbi	DDRB,DATA4
;		rjmp	prni_transfer

prni_transfer:
; read printer status
		sbrs	PrnFlags,PRN_STATUS_MODE
		rjmp	prni_pres
		mov	IntTemp,PrnBusy
		ori	IntTemp,(1<<DATA1)+(1<<CE1)+(1<<CE2)+(1<<KEY)
		out	PORTB,IntTemp
		sbi	DDRB,DATA4
		rjmp	ext_int1_end

; read the printer presence
prni_pres:	sbrs	PrnFlags,PRN_PRES_MODE
		rjmp	prni_datctr
		in	IntTemp,PINB
		com	IntTemp
		lsl	IntTemp			;move the PRN_ENABLE bit to
						;the IntTemp,DATA4
		andi	IntTemp,(1<<DATA4)
		ori	IntTemp,(1<<DATA1)+(1<<CE1)+(1<<CE2)+(1<<KEY)
		out	PORTB,IntTemp
		sbi	DDRB,DATA4
		rjmp	ext_int1_end

; printer data or control
prni_datctr:	sbrs	PrnFlags,PRN_DATA_MODE
		rjmp	ext_int1_end
		lsr	PrnWordHigh
		ror	PrnWordLow
		sbis	PINB,DATA4
		or	PrnWordHigh,Const0x80
		rjmp	ext_int1_end

; OP=0, read command code to the CmdCode register
bus_cmd:	lsr	CmdCode
		sbis	PINB,DATA4
		ori	CmdCode,0x20
		sbr	Flags,(1<<CMD_RECEIVED)
ext_int1_end:	out	SREG,IntSregSave
		reti


; IRQ0 handle routine called at the falling edge of the CE3 pulse
ext_int0:	nop
		nop
		nop
		nop
		sbic	PIND,CE3
		reti
		cbi	DDRB,DATA4
		cbi	PORTB,DATA4	;disable pull-ups on DATA4 pin
		in	IntSregSave,SREG
; command received ?
		sbrs	Flags,CMD_RECEIVED
		rjmp	ext_int0_tape
		rcall	cmd_eval
		cbr	Flags,(1<<CMD_RECEIVED)
		rjmp	ext_int0_end2
; tape data transfer ?
ext_int0_tape:	sbrc	Flags,TAPE_DATA_MODE
		sbr	Flags,(1<<TAPE_BUSY)
; printer data transfer ?
		sbrs	PrnFlags,PRN_DATA_MODE
		rjmp	ext_int0_end1
		ldi	IntTemp,(1<<DATA4)
		mov	PrnBusy,IntTemp
		lsr	PrnWordHigh
		ror	PrnWordLow
		sbrs	PrnFlags,PRN_CTRL_MODE
		rjmp	ext_int0_prn2
		mov	PrnControl,PrnWordLow
		sbr	PrnFlags,(1<<PRN_CTRL_RDY)
		rjmp	ext_int0_end1
ext_int0_prn2:	mov	PrnData,PrnWordLow
		sbr	PrnFlags,(1<<PRN_DATA_RDY)
ext_int0_end1:	cbr	Flags,(1<<TAPE_DATA_MODE)+(1<<TAPE_STATUS_MODE)
		cbr	PrnFlags,(1<<PRN_STATUS_MODE)+(1<<PRN_PRES_MODE)+(1<<PRN_DATA_MODE)
ext_int0_end2:	out	SREG,IntSregSave
		reti


; CmdCode evaluate
cmd_eval:	andi	CmdCode,0x3F
		cpi	CmdCode,0x00		;reset ?
		brne	cmd_eval_0
		clr	Flags
		clr	PrnFlags
		ret
cmd_eval_0:	sbrc	CmdCode,5
		rjmp	cmd_eval_10
; tape interface commands
		cpi	CmdCode,0x08		;read status ?
		brne	cmd_eval_3
		sbr	Flags,(1<<TAPE_STATUS_MODE)
		cbr	Flags,(1<<TAPE_DATA_MODE)
		ret
cmd_eval_3:	cpi	CmdCode,0x12		;transfer data ?
		brne	cmd_eval_4
		sbr	Flags,(1<<TAPE_DATA_MODE)
		cbr	Flags,(1<<TAPE_STATUS_MODE)
		ret
cmd_eval_4:	cpi	CmdCode,0x0F		;close ?
		brne	cmd_eval_6
		clr	Flags
		ret
cmd_eval_6:	cpi	CmdCode,0x0C		;wait ?
		brne	cmd_eval_7
		sbr	Flags,(1<<FORCE_BUSY)+(1<<TAPE_STATUS_MODE)
		cbr	Flags,(1<<TAPE_DATA_MODE)
		ldi	IntTemp,0x05		;CK/1024
		out	TCCR1B,IntTemp		;start the Timer1
		ret
; tape open commands only supported when TAPE_ENABLE high
cmd_eval_7:	sbis	PINB,TAPE_ENABLE
		rjmp	cmd_unknown		;unknown command
		cpi	CmdCode,0x05		;open for write ?
		brne	cmd_eval_8
		ldi	Flags,(1<<TAPE_ACTIVE)+(1<<TAPE_START)
		ret
cmd_eval_8:	cpi	CmdCode,0x06		;open for read ?
		brne	cmd_unknown		;unknown command
		ldi	Flags,(1<<TAPE_BUSY)+(1<<TAPE_LOAD)+(1<<TAPE_ACTIVE)+(1<<TAPE_START)
		ret
; printer commands
cmd_eval_10:	cpi	CmdCode,0x28		;read status ?
		brne	cmd_eval_11
		sbr	PrnFlags,(1<<PRN_STATUS_MODE)
		cbr	PrnFlags,(1<<PRN_PRES_MODE)+(1<<PRN_DATA_MODE)
		ret
cmd_eval_11:	cpi	CmdCode,0x33		;send control ?
		brne	cmd_eval_12
		sbr	PrnFlags,(1<<PRN_DATA_MODE)+(1<<PRN_CTRL_MODE)
		cbr	PrnFlags,(1<<PRN_PRES_MODE)+(1<<PRN_STATUS_MODE)
		ret
cmd_eval_12:	cpi	CmdCode,0x32		;send data ?
		brne	cmd_eval_13
		sbr	PrnFlags,(1<<PRN_DATA_MODE)
		cbr	PrnFlags,(1<<PRN_PRES_MODE)+(1<<PRN_CTRL_MODE)+(1<<PRN_STATUS_MODE)
		ret
cmd_eval_13:	cpi	CmdCode,0x30		;printer presence ?
		brne	cmd_unknown		;unknown command
		sbr	PrnFlags,(1<<PRN_PRES_MODE)
		cbr	PrnFlags,(1<<PRN_DATA_MODE)+(1<<PRN_STATUS_MODE)
cmd_unknown:	ret


start:		cli
		ldi	Temp1,0xDF
		out	SPL,Temp1
		ldi	Temp1,(1<<DATA1)+(1<<CE1)+(1<<CE2)+(1<<KEY)	;pull-ups
		out	PORTB,Temp1
		ldi	Temp1,0b00000000
		out	DDRB,Temp1
		ldi	Temp1,(1<<TXD)
		out	PORTD,Temp1
		out	DDRD,Temp1

		clr	Flags
		clr	PrnFlags
		ldi	Temp1,0x80
		mov	Const0x80,Temp1
		ldi	Temp1,(1<<DATA4)
		mov	PrnBusy,Temp1
		ldi	XL,low (TapeBuffer)
		ldi	XH,high (TapeBuffer)
		ldi	YL,low (TapeBuffer)
		ldi	YH,high (TapeBuffer)
		ldi	ZL,low (PrnBuffer)
		ldi	ZH,high (PrnBuffer)

; timer0 used to count the SCL frequency
		ldi	Temp1,0x02	;CK/8
		out	TCCR0,Temp1

; UART parameters: 9600 baud, 8-bit, interrupts disabled
		ldi	Temp1,63	;or 64 when fck=10MHz
		out	UBRR,Temp1
		ldi	Temp1,0x18	;Transmitter & Receiver Enable
		out	UCR,Temp1

; falling edge of INT0 and INT1 should generate an interrupt request
		ldi	Temp1,0x0A
		out	MCUCR,Temp1
		ldi	Temp1,0xC0	;external interrupt 0 and 1 enable
		out	GIMSK,Temp1
		sei

		rcall	i2c_reset
		rcall	printer_init

main_loop:	sbrc	Flags,TAPE_ACTIVE	;tape access request?
		rcall	tape
		rcall	print
		rjmp	main_loop


print:		rcall	printer_busy
		brcs	print_13
		brts	print_13
; printer ready to accept data
		sbrs	PrnFlags,PRN_BUF_FULL
		rjmp	print_3

; printing data from the PrnBuffer
		ldi	Temp1,PrnBuffer
		cp	Temp1,ZL
		brcc	print_2
		ld	Temp1,-z
		rcall	print_char
		ret
; all data from the PrnBuffer have been printed
print_2:	cbr	PrnFlags,(1<<PRN_BUF_FULL)
		clr	PrnBusy
		ret

; receiving data from the calculator
print_3:	sbrs	PrnFlags,PRN_CTRL_RDY
		rjmp	print_5
		mov	Temp1,PrnControl
		cpi	Temp1,0x1F
		brne	print_4
		ldi	Temp1,0x0A		;line feed
		st	z+,Temp1
		ldi	Temp1,0x0D		;carriage return
		st	z+,Temp1
		sbr	PrnFlags,(1<<PRN_BUF_FULL)
print_4:	mov	PrnCtrlCopy,PrnControl
		cbr	PrnFlags,(1<<PRN_CTRL_RDY)
		ret

print_5:	sbrs	PrnFlags,PRN_DATA_RDY	;any data to print received ?
		rjmp	print_11
		cpi	ZL,PrnBuffer+PRNBUFSIZE
		brcc	print_10
		mov	r0,PrnData
		mov	Temp1,PrnCtrlCopy
		cpi	Temp1,0x11		;FX-602 number ?
		brne	print_8
; an FX-602 number
		ldi	Temp1,0x0F
		and	r0,Temp1
		rcall	print_conv
		st	z+,Temp1
		mov	r0,PrnData
		swap	r0
		ldi	Temp1,0x0F
		and	r0,Temp1
; not an FX-602 number
print_8:	rcall	print_conv
		brcc	print_9
		st	z+,r0
print_9:	st	z+,Temp1
print_10:	cbr	PrnFlags,(1<<PRN_DATA_RDY)
print_11:	rcall	test_prn_buf
		brcc	print_12
		clr	PrnBusy
		ret
print_12:	rcall	prn_buf_adj
		sbr	PrnFlags,(1<<PRN_BUF_FULL)
print_end:	ret

print_13:	ldi	Temp1,(1<<DATA4)
		mov	PrnBusy,Temp1
		ret


; returns Carry clr when PrnBuffer full
test_prn_buf:	mov	Temp1,PrnCtrlCopy
		cpi	Temp1,0x11		;FX-602, numbers
		ldi	Temp1,PrnBuffer+16
		breq	test_prn_buf1
		ldi	Temp1,PrnBuffer+1
test_prn_buf1:	cp	ZL,Temp1
		ret


; converts the character code in R0 to ASCII,
; returns first ASCII character in Temp1,
; also optionally second ASCII character in R0 and Carry set
print_conv:	push	ZL
		push	ZH
		mov	Temp1,r0
		cpi	Temp1,0xFF	;cr/lf ?
		brne	print_conv_0
		ldi	Temp1,0x88	;cr/lf pair in the casio_strings table
		rjmp	print_conv_3
print_conv_0:	cpi	Temp1,0x80
		brcc	print_conv_1	;illegal character
		mov	Temp1,PrnCtrlCopy
		ldi	ZL,low (2*fx702toascii)
		ldi	ZH,high (2*fx702toascii)
		cpi	Temp1,0x29	;FX-702 mode
		breq	print_conv_2
		ldi	ZL,low (2*fx602toascii)
		ldi	ZH,high (2*fx602toascii)
		cpi	Temp1,0x10	;FX-602 alpha mode
		breq	print_conv_2
		ldi	ZL,low (2*fx602num)
		ldi	ZH,high (2*fx602num)
		cpi	Temp1,0x11	;FX-602 numbers
		breq	print_conv_2
; unrecognized character or control code
print_conv_1:	ldi	Temp1,'~'
		clc			;no more characters		
		rjmp	print_conv_end
; Z points to the character conversion table
print_conv_2:	add	ZL,r0
		ldi	Temp1,0
		adc	ZH,Temp1
		lpm
		clc
		mov	Temp1,r0
		tst	Temp1
		brpl	print_conv_end	;no more characters
; get a pair of characters
print_conv_3:	ldi	ZL,low (2*casio_strings-0x80)
		ldi	ZH,high (2*casio_strings-0x80)
		add	ZL,Temp1
		ldi	Temp1,0
		adc	ZH,Temp1
		lpm
		mov	Temp1,r0
		adiw	ZL,1
		lpm
		sec
print_conv_end:	pop	ZH
		pop	ZL
		ret


; adjust the PrnBuffer contents in the FX-602 number mode
prn_buf_adj:	mov	Temp1,PrnCtrlCopy
		cpi	Temp1,0x11		;FX-602 number ?
		brne	prn_buf_adj_3
		ldi	Temp1,' '
		sts	PrnBuffer+15,Temp1
		lds	r0,PrnBuffer		;second digit of the exponent
		cp	r0,Temp1
		brne	prn_buf_adj_1
		lds	r0,PrnBuffer+1		;first digit of the exponent
		cp	r0,Temp1
		breq	prn_buf_adj_2
prn_buf_adj_1:	ldi	Temp1,'E'
prn_buf_adj_2:	sts	PrnBuffer+3,Temp1
prn_buf_adj_3:	ret


; prepare data on the control/status port,
; returns Carry set when the parallel interface is not available
printer_init:	rcall	i2c_start
		ldi	DataByte,0x40
		sec
		rcall	i2c_data
		brcs	print_char_3
		ldi	DataByte,0b11101111	;SLCTIN low
		sec
		rcall	i2c_data
		rjmp	print_char_3


; read the PAR_BUSY input state to the T flag,
; returns Carry set when the parallel interface is not available
printer_busy:	rcall	i2c_start
		ldi	DataByte,0x41
		sec
		rcall	i2c_data
		brcs	print_char_3
		ldi	DataByte,0xFF
		sec				;NOACK
		rcall	i2c_data
		bst	DataByte,PAR_BUSY
		rcall	i2c_stop
		clc
		ret


; transmit the Temp1 byte through the parallel port,
; returns Carry set when the parallel interface is not available
print_char:	push	Temp1
; prepare data on the data port
		rcall	i2c_start
		ldi	DataByte,0x42
		sec
		rcall	i2c_data
		pop	DataByte
		brcs	print_char_3
		sec
		rcall	i2c_data
		brcs	print_char_3
		rcall	i2c_stop
; send a strobe pulse
		rcall	i2c_start
		ldi	DataByte,0x40
		sec
		rcall	i2c_data
		brcs	print_char_3
		ldi	DataByte,0b01101111	;SLCTIN and STROBE low
		sec
		rcall	i2c_data
		brcs	print_char_3
		ldi	DataByte,0b11101111	;SLCTIN low, STROBE high
		sec
		rcall	i2c_data
print_char_3:	in	r0,SREG
		rcall	i2c_stop
		out	SREG,r0
		ret


; mass storage handle
tape:		sbrs	Flags,TAPE_START
		rjmp	tape_1
		cbr	Flags,(1<<TAPE_START)
		rcall	stop_timer1
		clr	FlowContrChar
		mov	XL,YL		;clear the tape buffer
		ldi	Skipped,5
		clr	TapeCounter
tape_1:		sbrs	Flags,TAPE_LOAD
		rjmp	serial_write

; LOAD mode, receiving data from the serial port
serial_read:
; read a character from the serial port, if available
		sbis	USR,7		;UART Receive Complete
		rjmp	serial_read_1
		in	Temp1,UDR
; skip any XON/XOFF characters
		cpi	Temp1,XON
		breq	serial_read_9
		cpi	Temp1,XOFF
		breq	serial_read_9
		rcall	store_tape_buf		;store the received character
; calculate how many characters are there in the tape buffer, and emit the
; XON/XOFF character when necessary (and when the transmitter is free)
serial_read_1:	ldi	Temp1,XON
		mov	r0,XL
		sub	r0,YL
		breq	serial_read_3
		brcc	serial_read_2
		ldi	Temp1,TAPEBUFSIZE
		add	r0,Temp1
serial_read_2:	ldi	Temp1,TAPEBUFSIZE/2	;limit arbitraly chosen
		cp	r0,Temp1
		brcs	serial_read_4
		ldi	Temp1,XOFF
serial_read_3:	sbis	USR,5			;TX Data Register Empty
		rjmp	serial_read_4
		cp	FlowContrChar,Temp1
		breq	serial_read_4		;character already transmitted
		mov	FlowContrChar,Temp1
		out	UDR,Temp1
; delay ?
serial_read_4:	in	Temp1,TCCR1B
		tst	Temp1			;Timer1 counting ?
		breq	serial_read_5		;branch if not
		in	Temp1,TCNT1L
		in	Temp1,TCNT1H
		tst	Temp1
		brmi	serial_read_9
		rcall	stop_timer1
; is the calculator ready to accept data ?
serial_read_5:	sbrs	Flags,TAPE_BUSY
		rjmp	serial_read_9		;previous data not sent yet
; is there any character in the tape buffer ?
		cp	XL,YL
		breq	serial_read_9		;tape buffer empty
		rcall	fetch_tape_buf		;fetch the character
; substract the offset and validate the character
		subi	Temp1,0x30
		brcs	serial_read_8
		cpi	Temp1,0x40
		brcc	serial_read_8
; shift right the 6-bit character into the DataWord variable
		rol	DataWordHigh
		rol	DataWordHigh
		mov	DataWordLow,DataWordHigh
		mov	DataWordHigh,Temp1
		rol	DataWordHigh
		ldi	Temp1,0x07
		or	DataWordLow,Temp1
		inc	TapeCounter
		mov	Temp1,TapeCounter
		cpi	Temp1,2			;already 2 characters shifted?
		brcs	serial_read_9
; the DataWord 16-bit variable contains now a valid 12-bit word on bits 14...3
; skip first 5 received data words
		tst	Skipped
		breq	serial_read_7
		dec	Skipped
		rjmp	serial_read_8
; prepare data to send to the calculator in DataWordLow, DataWordHigh and
; WrPortB to save time in the IRQ1 handle routine
serial_read_7:	com	DataWordLow
		com	DataWordHigh
		ldi	WrPortB,(1<<DATA1)+(1<<CE1)+(1<<CE2)+(1<<KEY)
					;initial data to be written to port B
		cbr	Flags,(1<<TAPE_BUSY)	;data ready
serial_read_8:	clr	TapeCounter
serial_read_9:	ret

; SAVE mode, sending data to the serial port
serial_write:
; is there any character in the tape buffer ?
		cp	XL,YL
		breq	serial_write_3		;tape buffer empty
; read the XON/XOFF character, if available
		sbis	USR,7			;UART Receive Complete
		rjmp	serial_write_2
		in	Temp1,UDR
		cpi	Temp1,XON
		breq	serial_write_1
		cpi	Temp1,XOFF
		brne	serial_write_2
serial_write_1:	mov	FlowContrChar,Temp1
; transmission allowed ?
serial_write_2:	mov	Temp1,FlowContrChar
		cpi	Temp1,XOFF
		breq	serial_write_5
		sbis	USR,5			;TX Data Register Empty ?
		rjmp	serial_write_5
; fetch a character from the tape buffer and transmit it through the serial port
		rcall	fetch_tape_buf
		out	UDR,Temp1
		rjmp	serial_write_5

; any data received from the calculator ?
serial_write_3:	sbrs	Flags,TAPE_BUSY
		rjmp	serial_write_5
; insert the CR/LF sequence every 32 data words
		mov	Temp1,TapeCounter
		andi	Temp1,0x1F
		brne	serial_write_4
		ldi	Temp1,0x0D
		rcall	store_tape_buf
		ldi	Temp1,0x0A
		rcall	store_tape_buf
; the least significant character
serial_write_4:	mov	Temp1,DataWordLow
		andi	Temp1,0x3F
		subi	Temp1,0xD0		;add offset 0x30
		rcall	store_tape_buf
; the most significant character
		mov	Temp1,DataWordLow
		eor	Temp1,DataWordHigh
		andi	Temp1,0xC0
		eor	Temp1,DataWordHigh
		swap	Temp1
		lsr	Temp1
		lsr	Temp1
		subi	Temp1,0xD0		;add offset 0x30
		rcall	store_tape_buf
		inc	TapeCounter
		cbr	Flags,(1<<TAPE_BUSY)	;data sent
serial_write_5:	ret


; store the character in Temp1 to the tape buffer
store_tape_buf:	st	x+,Temp1
		cpi	XL,TapeBuffer+TAPEBUFSIZE
		brcs	store_tape_bu1
		subi	XL,TAPEBUFSIZE		;wrap the write pointer
store_tape_bu1:	ret


; fetch the character from the tape buffer to Temp1
fetch_tape_buf:	ld	Temp1,y+
		cpi	YL,TapeBuffer+TAPEBUFSIZE
		brcs	fetch_tape_bu1
		subi	YL,TAPEBUFSIZE		;wrap the read pointer
fetch_tape_bu1:	ret


; stop the Timer1 and preset it for a 150ms delay
stop_timer1:	cbr	Flags,(1<<FORCE_BUSY)
		ldi	Temp1,0x00
		out	TCCR1B,Temp1
		ldi	Temp1,high (-1500)
		out	TCNT1H,Temp1
		ldi	Temp1,low (-1500)
		out	TCNT1L,Temp1
		ret


; wait ca. 5us
wait_5us:	ldi	Temp1,256-7	;7 * 0.8us = 5.6us
		out	TCNT0,Temp1
wait_2:		in	Temp1,TCNT0
		sbrc	Temp1,7
		rjmp	wait_2
		ret


; *** I2C bus routines ***

; send byte DataByte and read ACK to Carry (on input Carry set required),
; or read byte DataByte and set ACK according to Carry (on input DataByte=0xFF
; required)
i2c_data:	ldi	Counter,9	;bit counter
i2c_data_loop:	rol	DataByte
; copy Carry to SDA
		brcc	i2c_data_1
		cbi	DDRD,SDA	;SDA=1
		rjmp	i2c_data_2
i2c_data_1:	sbi	DDRD,SDA	;SDA=0
i2c_data_2:	rcall	wait_5us
; SCL=1
		cbi	DDRD,SCL	;SCL=1
; read SDA to Carry
		sbis	PIND,SDA
		rjmp	i2c_data_3
		sec
		rjmp	i2c_data_4
i2c_data_3:	clc
i2c_data_4:	rcall	wait_5us
; SCL=0
		sbi	DDRD,SCL	;SCL=0
; next bit
		dec	Counter
		brne	i2c_data_loop
		ret

; send start condition
i2c_start:	rcall	wait_5us
		sbi	DDRD,SCL	;SCL=0
		rcall	wait_5us
		cbi	DDRD,SDA	;SDA=1
		rcall	wait_5us
		cbi	DDRD,SCL	;SCL=1
i2c_start_1:	rcall	wait_5us
		sbi	DDRD,SDA	;SDA=0
		rcall	wait_5us
		sbi	DDRD,SCL	;SCL=0
		ret

; send stop condition
i2c_stop:	rcall	wait_5us
		sbi	DDRD,SCL	;SCL=0
		rcall	wait_5us
		sbi	DDRD,SDA	;SDA=0
		rcall	wait_5us
		cbi	DDRD,SCL	;SCL=1
		rcall	wait_5us
		cbi	DDRD,SDA	;SDA=1
		ret

; I2C bus reset
i2c_reset:	cbi	DDRD,SDA	;SDA=1
		rcall	wait_5us
		cbi	DDRD,SCL	;SCL=1
		rcall	wait_5us
		ldi	Counter,9	;up to 9 clock pulses
i2c_reset_1:	sbic	PIND,SDA	;look for SDA high while SCL is high
		rjmp	i2c_reset_2
		rcall	wait_5us
		sbi	DDRD,SCL	;SCL=0
		rcall	wait_5us
		cbi	DDRD,SCL	;SCL=1
		dec	Counter
		brne	i2c_reset_1
i2c_reset_2:	rcall	i2c_start_1
		rcall	i2c_stop
		ret
